@@ -3,19 +3,23 @@
 
 from abc import ABC, abstractmethod
 import time
+import os
 from operator import attrgetter, methodcaller
 from collections import defaultdict
-from typing import List, Tuple, Iterator, Optional
+from typing import List, Tuple, Iterator, Optional, Type
 import logging
 from concurrent.futures import ThreadPoolExecutor
 
 from runtype import dataclass
 
-from .sql import Select, Checksum, Compare, DbPath, DbKey, DbTime, Count, TableName, Time, Min, Max
+from .sql import Select, Checksum, Compare, DbPath, DbKey, DbTime, Count, TableName, Time, Min, Max, Value
+from .utils import safezip, split_space
 from .databases.base import Database
 from .databases.database_types import (
+    ArithUUID,
     NumericType,
     PrecisionType,
+    StringType,
     UnknownColType,
     Schema,
     Schema_CaseInsensitive,
@@ -25,22 +29,11 @@ from .databases.database_types import (
 logger = logging.getLogger("diff_tables")
 
 RECOMMENDED_CHECKSUM_DURATION = 10
-
+BENCHMARK = os.environ.get("BENCHMARK", False)
 DEFAULT_BISECTION_THRESHOLD = 1024 * 16
 DEFAULT_BISECTION_FACTOR = 32
 
 
-def safezip(*args):
-    "zip but makes sure all sequences are the same length"
-    assert len(set(map(len, args))) == 1
-    return zip(*args)
-
-
-def split_space(start, end, count):
-    size = end - start
-    return list(range(start, end, (size + 1) // (count + 1)))[1 : count + 1]
-
-
 @dataclass(frozen=False)
 class TableSegment:
     """Signifies a segment of rows (and selected columns) within a table
@@ -86,25 +79,34 @@ class TableSegment:
         if self.min_update is not None and self.max_update is not None and self.min_update >= self.max_update:
             raise ValueError("Error: min_update expected to be smaller than max_update!")
 
-    @property
-    def _key_column(self):
-        return self._quote_column(self.key_column)
-
     @property
     def _update_column(self):
         return self._quote_column(self.update_column)
 
-    def _quote_column(self, c):
+    def _quote_column(self, c: str) -> str:
         if self._schema:
-            c = self._schema.get_key(c)
+            c = self._schema.get_key(c)  # Get the actual name. Might be case-insensitive.
         return self.database.quote(c)
 
+    def _normalize_column(self, name: str, template: str = None) -> str:
+        if not self._schema:
+            raise RuntimeError(
+                "Cannot compile query when the schema is unknown. Please use TableSegment.with_schema()."
+            )
+
+        col = self._quote_column(name)
+        if template is not None:
+            col = template % col  # Apply template using Python's string formatting
+
+        return self.database.normalize_value_by_type(col, self._schema[name])
+
     def with_schema(self) -> "TableSegment":
         "Queries the table schema from the database, and returns a new instance of TableSegmentWithSchema."
         if self._schema:
             return self
 
         schema = self.database.query_table_schema(self.table_path, self._relevant_columns)
+        logger.debug(f"[{self.database.name}] Schema = {schema}")
 
         schema_inst: Schema
         if self.case_sensitive:
@@ -121,9 +123,9 @@ class TableSegment:
 
     def _make_key_range(self):
         if self.min_key is not None:
-            yield Compare("<=", str(self.min_key), self._key_column)
+            yield Compare("<=", Value(self.min_key), self._quote_column(self.key_column))
         if self.max_key is not None:
-            yield Compare("<", self._key_column, str(self.max_key))
+            yield Compare("<", self._quote_column(self.key_column), Value(self.max_key))
 
     def _make_update_range(self):
         if self.min_update is not None:
@@ -133,7 +135,7 @@ class TableSegment:
 
     def _make_select(self, *, table=None, columns=None, where=None, group_by=None, order_by=None):
         if columns is None:
-            columns = [self._key_column]
+            columns = [self._normalize_column(self.key_column)]
         where = list(self._make_key_range()) + list(self._make_update_range()) + ([] if where is None else [where])
         order_by = None if order_by is None else [order_by]
         return Select(
@@ -152,6 +154,11 @@ class TableSegment:
     def choose_checkpoints(self, count: int) -> List[DbKey]:
         "Suggests a bunch of evenly-spaced checkpoints to split by (not including start, end)"
         assert self.is_bounded
+        if isinstance(self.min_key, ArithUUID):
+            checkpoints = split_space(self.min_key.int, self.max_key.int, count)
+            assert isinstance(self.max_key, ArithUUID)
+            return [ArithUUID(int=i) for i in checkpoints]
+
         return split_space(self.min_key, self.max_key, count)
 
     def segment_by_checkpoints(self, checkpoints: List[DbKey]) -> List["TableSegment"]:
@@ -185,14 +192,7 @@ class TableSegment:
 
     @property
     def _relevant_columns_repr(self) -> List[str]:
-        if not self._schema:
-            raise RuntimeError(
-                "Cannot compile query when the schema is unknown. Please use TableSegment.with_schema()."
-            )
-        return [
-            self.database.normalize_value_by_type(self._quote_column(c), self._schema[c])
-            for c in self._relevant_columns
-        ]
+        return [self._normalize_column(c) for c in self._relevant_columns]
 
     def count(self) -> Tuple[int, int]:
         """Count how many rows are in the segment, in one pass."""
@@ -215,7 +215,13 @@ class TableSegment:
 
     def query_key_range(self) -> Tuple[int, int]:
         """Query database for minimum and maximum key. This is used for setting the initial bounds."""
-        select = self._make_select(columns=[Min(self._key_column), Max(self._key_column)])
+        # Normalizes the result (needed for UUIDs) after the min/max computation
+        select = self._make_select(
+            columns=[
+                self._normalize_column(self.key_column, "min(%s)"),
+                self._normalize_column(self.key_column, "max(%s)"),
+            ]
+        )
         min_key, max_key = self.database.query(select, tuple)
 
         if min_key is None or max_key is None:
@@ -297,9 +303,16 @@ class TableDiffer:
         key_ranges = self._threaded_call("query_key_range", [table1, table2])
         mins, maxs = zip(*key_ranges)
 
+        key_type = table1._schema[table1.key_column]
+        key_type2 = table2._schema[table2.key_column]
+        assert key_type.python_type is key_type2.python_type
+
         # We add 1 because our ranges are exclusive of the end (like in Python)
-        min_key = min(map(int, mins))
-        max_key = max(map(int, maxs)) + 1
+        try:
+            min_key = min(map(key_type.python_type, mins))
+            max_key = max(map(key_type.python_type, maxs)) + 1
+        except (TypeError, ValueError) as e:
+            raise type(e)(f"Cannot apply {key_type} to {mins}, {maxs}.") from e
 
         table1 = table1.new(min_key=min_key, max_key=max_key)
         table2 = table2.new(min_key=min_key, max_key=max_key)
@@ -324,7 +337,7 @@ class TableDiffer:
             col2 = table2._schema[c]
             if isinstance(col1, PrecisionType):
                 if not isinstance(col2, PrecisionType):
-                    raise TypeError(f"Incompatible types for column {c}:  {col1} <-> {col2}")
+                    raise TypeError(f"Incompatible types for column '{c}':  {col1} <-> {col2}")
 
                 lowest = min(col1, col2, key=attrgetter("precision"))
 
@@ -336,7 +349,7 @@ class TableDiffer:
 
             elif isinstance(col1, NumericType):
                 if not isinstance(col2, NumericType):
-                    raise TypeError(f"Incompatible types for column {c}:  {col1} <-> {col2}")
+                    raise TypeError(f"Incompatible types for column '{c}':  {col1} <-> {col2}")
 
                 lowest = min(col1, col2, key=attrgetter("precision"))
 
@@ -346,12 +359,16 @@ class TableDiffer:
                 table1._schema[c] = col1.replace(precision=lowest.precision)
                 table2._schema[c] = col2.replace(precision=lowest.precision)
 
+            elif isinstance(col1, StringType):
+                if not isinstance(col2, StringType):
+                    raise TypeError(f"Incompatible types for column '{c}':  {col1} <-> {col2}")
+
         for t in [table1, table2]:
             for c in t._relevant_columns:
                 ctype = t._schema[c]
-                if isinstance(ctype, UnknownColType):
-                    logger.warn(
-                        f"[{t.database.name}] Column '{c}' of type '{ctype.text}' has no compatibility handling. "
+                if not ctype.supported:
+                    logger.warning(
+                        f"[{t.database.name}] Column '{c}' of type '{ctype}' has no compatibility handling. "
                         "If encoding/formatting differs between databases, it may result in false positives."
                     )
 
@@ -403,6 +420,16 @@ class TableDiffer:
             f"size: {table2.max_key-table1.min_key}"
         )
 
+        # When benchmarking, we want the ability to skip checksumming. This
+        # allows us to download all rows for comparison in performance. By
+        # default, data-diff will checksum the section first (when it's below
+        # the threshold) and _then_ download it.
+        if BENCHMARK:
+            max_rows_from_keys = max(table1.max_key - table1.min_key, table2.max_key - table2.min_key)
+            if max_rows_from_keys < self.bisection_threshold:
+                yield from self._bisect_and_diff_tables(table1, table2, level=level, max_rows=max_rows_from_keys)
+                return
+
         (count1, checksum1), (count2, checksum2) = self._threaded_call("count_and_checksum", [table1, table2])
 
         if count1 == 0 and count2 == 0:
